import re as __re


def __indent(__str: str):
    return len(__str) - len(__str.lstrip())


def __first_letter_match(__str: str):
    return __re.search("[a-zA-Z]", __str)


def __from_match_start(__str: str, __match: __re.Match):
    return __str[__match.start() if __match else 0 :]


def __from_first_letter(__str: str):
    match = __first_letter_match(__str)
    return __from_match_start(__str, match)


def __safe_merge(__lhs:dict,__rhs:dict):
    result = __lhs.copy()
    for key,value in __rhs.items():
        if key in __lhs:
            lvalue = __lhs.get(key)
            if lvalue != value:
                raise Exception(f"ERROR: CONFLICT OF DICT KEYS:: {str(key)} : {lvalue} : {value}")
        result[key] = value
    return result


def from_first_letter(__str: str):
    return __from_first_letter(__str)


def get_indent(__str: str):
    return __indent(__str)


def safe_merge(__lhs:dict,__rhs:dict):
    return __safe_merge(__lhs,__rhs)


from enum import Enum
from pathlib import Path
from typing import List,Dict
from .util import get_indent, from_first_letter


def path_exists(__path: str):
    return Path(__path).exists()


def read_text(__path: str):
    return Path(__path).read_text()


def write_text(path: str, data: str):
    Path(path).write_text(data=data)


class EnvType(Enum):

    DOTENV = 1
    DOCKERFILE = 2
    DOCKER_COMPOSE = 3
    NONE = 4


def __env_dict_from_lines(__env_lines: List[str]):
    env_lines = __env_lines.copy()
    env_lines.sort()
    env_kv_pairs = [line.split("=", 1) for line in env_lines if "=" in line]
    return {key: value for (key, value) in env_kv_pairs}


def docker_compose_lines(__docker_compose_src: str):
    indent, results = None, []
    for line in __docker_compose_src.splitlines(keepends=False):
        line_indent = get_indent(line)
        if indent is None:
            if line.strip().endswith("environment:"):
                indent = line_indent
        elif line_indent > indent:
            ffl = from_first_letter(line)
            if "=" in ffl:
                results.append(ffl)
            elif ":" in ffl:
                lhs, rhs = ffl.split(":", 1)
                ffl = f"{lhs.rstrip()}={rhs.lstrip()}"
                results.append(ffl)
        else:
            indent = None
    return results


def dotenv_lines(__src: str):
    return [line for line in __src.splitlines(keepends=False) if "=" in line]


def dockerfile_lines(__src: str):
    return [
        from_first_letter(line.removeprefix("ENV"))
        for line in __src.splitlines(keepends=False)
        if line.lstrip().startswith("ENV") and "=" in line
    ]


def docker_compose_dict(__docker_compose_src: str):
    return __env_dict_from_lines(docker_compose_lines(__docker_compose_src))


def dockerfile_dict(__src: str):
    return __env_dict_from_lines(dockerfile_lines(__src))


def dotenv_dict(__dotenv_src: str):
    return __env_dict_from_lines(dotenv_lines(__dotenv_src))


def env_dict(path: str = None, src: str = None, env: EnvType = EnvType.NONE):
    if src is not None:
        if env == EnvType.DOCKER_COMPOSE:
            result = docker_compose_dict(src)
        elif env == EnvType.DOCKERFILE:
            result = dockerfile_dict(src)
        elif env == EnvType.DOTENV:
            result = dotenv_dict(src)
        else:
            raise Exception(f"unrecognized env_type: {str(env)}")

    else:
        if path is None:
            raise Exception("no param for env_dict")
        else:
            src = read_text(path)
        result = env_dict(src=src, env=env)
    return result


def form_dotenv_src(__dict: Dict[str, str]):
    env_lines = [f"{key}={value}" for key, value in __dict.items()]
    env_lines.sort()
    return "\n".join(env_lines) if env_lines else ""


def form_python_env_src(env: Dict[str, str], dotenv_path: str = ".env",python_path:str="env.py"):

    field_src = lambda key: f"    {key}:Optional[str] = Field(default=None)"
    keys = list(env.keys())
    keys.sort()
    env_fields_src = "\n".join(map(field_src, keys))
    env_init_src = "env = Env(**{" + "key:os.environ.get(key) for key in Env.keys()})\n"

    result = f"""
import os
from typing import Optional
from dotenv import load_dotenv
from pydantic import BaseModel, Field

load_dotenv('{dotenv_path}')

class Env(BaseModel):

    @classmethod
    def keys(cls):
        return cls.__fields__.keys()

{env_fields_src}

{env_init_src}

    """

    if path_exists(python_path):
        env_src = read_text(python_path)
        if env_init_src in env_src:
            env_src = env_init_src.split(env_init_src,1)[1]
            result += f"\n{env_src}"
    return result


def form_dockerfile_env_src(
    __dict: Dict[str, str], dockerfile_path: str = "Dockerfile"
):
    dockerfile_lines = []
    if path_exists(dockerfile_path):
        dockerfile_env_src = read_text(dockerfile_path)
        dockerfile_lines = dockerfile_env_src.splitlines(keepends=False)
    dockerfile_lines = [line for line in dockerfile_lines if not line.startswith("ENV")]
    endline = ""
    if dockerfile_lines:
        endline = dockerfile_lines[-1]
        dockerfile_lines = dockerfile_lines[:-1]
    env_lines = [f"ENV {key}={value}" for key, value in __dict.items()]
    result = dockerfile_lines + env_lines + [endline]
    return "\n".join(result)



from . import core
from . import util

def docker_compose_dict(path: str):
    return core.env_dict(path=path, env=core.EnvType.DOCKER_COMPOSE)


def dotenv_dict(path: str):
    return core.env_dict(path=path, env=core.EnvType.DOTENV)


def dockerfile_dict(path: str):
    return core.env_dict(path=path, env=core.EnvType.DOCKERFILE)



def refresh_envs(
    docker_compose_path: str = "docker-compose.yml",
    dotenv_path: str = ".env",
    dockerfile_path: str = "Dockerfile",
    python_env_path: str = "env.py",
):

    docker_env = {}
    if core.path_exists(docker_compose_path):
        docker_compose_env = docker_compose_dict(path=docker_compose_path)
        docker_env = util.safe_merge(docker_env,docker_compose_env)
    if core.path_exists(dockerfile_path):
        dockerfile_env = dockerfile_dict(dockerfile_path)
        docker_env = util.safe_merge(docker_env,dockerfile_env)
    if core.path_exists(dotenv_path):
        dotenv_env = dotenv_dict(dotenv_path)
        docker_env = util.safe_merge(docker_env,dotenv_env)

    if docker_env:
        dotenv_src = core.form_dotenv_src(docker_env)
        python_src = core.form_python_env_src(docker_env, dotenv_path)
        dockerfile_src = core.form_dockerfile_env_src(docker_env, dockerfile_path)
        core.write_text(dotenv_path, dotenv_src)
        core.write_text(python_env_path, python_src)
        core.write_text(dockerfile_path, dockerfile_src)
